"""model: Machine learning models for class prediction with Snekmer.

author: @christinehc

"""
# imports
import pandas as pd
from typing import Any, Dict, List, Optional
from ._version import __version__
from .vectorize import KmerBasis
from numpy.typing import NDArray
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.linear_model import LogisticRegression  # LogisticRegressionCV
from sklearn.model_selection import GridSearchCV, cross_validate
from sklearn.pipeline import make_pipeline, Pipeline
from sklearn.svm import SVC

# define default gridsearch param dict
MODEL_PARAMS = {
    #     'scaler__n': [None, 100],
    "clf__class_weight": ["balanced"],
    "clf__min_samples_split": [2, 5, 10, 15, 20, 40],
    "clf__min_samples_leaf": [2, 5, 10, 15, 20],
    "clf__max_depth": [3, 10, 25],
}

NAME2MODEL = {
    "decisiontree": DecisionTreeClassifier,
    "randomforest": RandomForestClassifier,
    "adaboost": AdaBoostClassifier,
    "logistic": LogisticRegression,
    "svc": SVC,
}


# classification models for protein families
class SnekmerModel(BaseEstimator):
    """Classify a protein family using kmer vectors as input.

    Attributes
    ----------
    scaler :  KmerScaler object or None
        Scaler object; if None, feature vectors are not scaled
    params : dict
        Parameter dictionary for hyperparameter gridsearch.
    model : str (default: 'decisiontree')
        String identifier for classifier model.
        Mappings to Classifier object are:
            {
            'decisiontree': DecisionTreeClassifier,
            'randomforest': RandomForestClassifier,
            'adaboost': AdaBoostClassifier
            }
    step_name : str (default: 'clf')
        Optional custom name for classifier pipeline step.
    pipeline : sklearn.pipeline.Pipeline object
        Pipeline object for preprocessing and modeling.
    search : sklearn.model_selection.GridSearchCV object
        GridSearchCV object for hyperparameter searching.

    """

    def __init__(
        self,
        scaler: Optional[Any] = None,
        model: str = "logistic",
        model_params: Optional[Dict[Any, Any]] = {},
        params: Dict[str, List] = MODEL_PARAMS,
        step_name="clf",
    ):
        """Initialize KmerModel object.

        Parameters
        ----------
        scaler :  KmerScaler object or None (default: None)
            Scaler object; if None, does not scale feature vectors
        model : Classifier object (default: DecisionTreeClassifier())
            Type of classification model.
        params : dict (default: MODEL_PARAMS)
            Parameter dictionary for hyperparameter gridsearch.
        step_name : str (default: 'clf')
            Optional custom name for classifier pipeline step.

        """
        # always compute svc probability
        if model == "svc":
            model_params["probability"] = True

        self.scaler = scaler
        self.params = params
        self.model = NAME2MODEL[model](**model_params)
        self.step_name = step_name
        self.snekmer_version = __version__

    def fit(
        self, X: NDArray, y: NDArray, gridsearch: bool = False, verbose: bool = True
    ):
        """Train model using gridsearch-tuned hyperparameters.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            Target values (class labels) as integers or strings.
        scores : array-like of shape (n_features,) or (n_features, n_outputs)
            NOT IMPLEMENTED YET-- for KmerScaler integration.
        gridsearch: bool (default: False)
            If True, uses grid search to determine hyperparameters.
        verbose : bool (default: True)
            If True, prints best gridsearch CV results to console.
            Only active if `gridsearch=True`.

        Returns
        -------
        None
            Fits estimator.

        """
        # define pipeline and gridsearch
        if gridsearch:
            self.pipeline = Pipeline(
                steps=[("scaler", self.scaler), (self.step_name, self.model)]
            )
            self.search = GridSearchCV(self.pipeline, self.params)

            # use gridsearch on training data to find best parameter set
            self.search.fit(X, y)
            if verbose:
                print(
                    "best mean cross-validation score: {:.3f}".format(
                        self.search.best_score_
                    )
                )
                print("best parameters:", self.search.best_params_)

        # fit model with best gridsearch parameters
        #         self.scaler.fit(scores)
        self.model.fit(X, y)
        return self

    def score(self, X: NDArray, y: NDArray):
        """Score model on test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            Predicted classes, or predict values.

        Returns
        -------
        float
            Model prediction accuracy on test set.

        """
        return self.search.score(X, y)

    def predict(self, X: NDArray):
        """Assign classification based on new input vector.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            New samples.

        Returns
        -------
        int
            (0,1) classification result

        """
        return self.model.predict(X)


class SnekmerModelCV:
    def __init__(self, model: str, cv: int = 5):
        self.model = model
        self.cv = cv

    def train(
        self,
        X: NDArray,
        y: NDArray,
        i: NDArray,
        label: str = "label",
        random_state: Optional[int] = None,
    ):
        """_summary_

        Parameters
        ----------
        X : NDArray
            _description_
        y : NDArray
            _description_
        i : NDArray
            _description_
        label : str, optional
            _description_, by default "label"
        random_state : Optional[int], optional
            _description_, by default None

        Raises
        ------
        ValueError
            _description_

        """
        # set and format input and label arrays; initialize model objects
        cols = [label, "alphabet_name", "k", "scoring"]

        results = {col: [] for col in cols + ["score", "cv_split"]}
        X, y = {i: {} for i in range(self.cv)}, {i: {} for i in range(self.cv)}
        for n in range(self.cv):

            # remove score cols that were generated from full dataset
            unscored_cols = [col for col in list(data.columns) if "_score" not in col]

            # filter data by training data per CV split
            i_train = data[data[f"train_cv-{n + 1:02d}"]].index
            i_test = data[~data[f"train_cv-{n + 1:02d}"]].index
            df_train = data.iloc[i_train][unscored_cols].reset_index(drop=True)
            df_test = data.iloc[i_test][unscored_cols].reset_index(drop=True)
            df_train_labels = [
                True if value == family else False for value in df_train[label]
            ]
            df_test_labels = [
                True if value == family else False for value in df_test[label]
            ]

            # score kmers separately per split
            scorer = skm.score.KmerScorer()
            scorer.fit(
                list(kmer.kmer_set.kmers),
                df_train,
                family,
                label_col=label,
                vec_col="sequence_vector",
                **config["score"]["scaler_kwargs"],
            )

            # append scored sequences to dataframe
            df_train = df_train.merge(
                pd.DataFrame(scorer.scores["sample"]), left_index=True, right_index=True
            )
            if df_train.empty:
                raise ValueError("Blank df")
            df_test = df_test.merge(
                pd.DataFrame(
                    scorer.predict(
                        skm.utils.to_feature_matrix(df_test["sequence_vector"]),
                        list(kmer.kmer_set.kmers),
                    )
                ),
                left_index=True,
                right_index=True,
            ).rename(columns={0: f"{family}_score"})

            # save score loadings
            scores = (
                pd.DataFrame(scorer.probabilities, index=scorer.kmers.basis)
                .reset_index()
                .rename(columns={"index": "kmer"})
            )

            # save X,y array data for plot
            X[n]["train"] = df_train[f"{family}_score"].values.reshape(-1, 1)
            y[n]["train"] = le.transform(df_train_labels).ravel()

            X[n]["test"] = df_test[f"{family}_score"].values.reshape(-1, 1)
            y[n]["test"] = le.transform(df_test_labels).ravel()

        # collate ROC-AUC results
        results[label] += [family] * cv
        results["alphabet_name"] += [alphabet_name.lower()] * cv
        results["k"] += [config["k"]] * cv
        results["scoring"] += ["auc_roc"] * cv
        results["score"] += auc_rocs
        results["cv_split"] += [i + 1 for i in range(cv)]



"""report: Snekmer report generation code.

author: @abbyjerger, @christinehc

"""
from jinja2 import Environment, PackageLoader
import os

env = Environment(loader=PackageLoader("snekmer"), auto_reload=False)

TEMPLATES = {
    "cluster": "cluster_template.html",
    "model": "model_template.html",
    "search": "search_template.html",
}


def correct_rel_path(filepath: str, out_dir: str = "output") -> str:
    """Correct relative file paths for Jinja.

    Parameters
    ----------
    filepath : str
        /path/to/file
    out_dir : str, optional
        /path/to/output, by default "output"

    Returns
    -------
    str
        /path/to/file
    """
    # force paths to use OS separator
    filepath = filepath.replace("/", os.sep)

    filepath = filepath.split(os.sep)
    out_dir = out_dir.split(os.sep)
    for d in out_dir:
        try:
            filepath.remove(d)
        except ValueError:  # if output dirs not found, skip
            pass
    return os.sep.join(filepath)


# cluster can use just this function since the image outputs are fixed at 3
# model and search also eventually call this function
def create_report(template_vars, template: str, report_file_name: str):
    """Create Snekmer reports.

    Parameters
    ----------
    template_vars : dict
        Variables defined in Jinja template
    template : str
        Name of template ("cluster", "model", or "search")
    report_file_name : str
        /path/to/report_file.html

    Returns
    -------
    None
        Creates file and exits.

    """
    # look in this file folder for the templates
    template = env.get_template(TEMPLATES[template])
    html = template.render(template_vars)
    with open(report_file_name, "w") as f:
        f.write(html)


# model
def create_report_many_images(
    image_path: str,
    table_path: str,
    rep_vars: dict,
    template: str,
    report_file_name: str,
):
    """Create report for Snekmer model mode.

    Parameters
    ----------
    image_path : str
        /path/to/images/
    table_path : str
        /path/to/tables/
    rep_vars : dict
        Variables defined in Jinja template
    template : str
        Name of template ("cluster", "model", or "search")
    report_file_name : str
        /path/to/report_file.html

    Returns
    -------
    None
        Creates file and exits.

    """
    image_list, table_list = list(), list()
    # collate images
    for root, dirs, files in os.walk(image_path):
        for file in files:
            if file.endswith(".png"):
                image_list.append(os.path.join(root, file))

    # collate tables
    for root, dirs, files in os.walk(table_path):
        for file in files:
            if file.endswith(".csv"):
                table_list.append(os.path.join(root, file))
    rep_vars["images"] = [correct_rel_path(f) for f in image_list]
    rep_vars["tables"] = [correct_rel_path(f) for f in table_list]

    create_report(rep_vars, template, report_file_name)


def create_report_many_csvs(
    path: str, rep_vars: dict, template: str, report_file_name: str
):
    # create lists with directory names and file names
    file_list = []
    dirs_list = []
    for root, dirs, files in os.walk(path):
        for d in dirs:
            dirs_list.append(d)
        for file in files:
            if file.endswith(".csv"):
                file_list.append(os.path.join(root, file))

    # combine both lists into one, in the correct order to print in the jinja template
    ordered = []
    for i in dirs_list:
        ordered.append(i)
        for j in file_list:
            if os.path.split(os.path.split(j)[0])[1] == i:
                ordered.append(j)

    rep_vars['dirs_list'] = dirs_list
    rep_vars['ordered'] = [correct_rel_path(f) for f in ordered]
    create_report(rep_vars, template, report_file_name)



"""vectorize: Create kmer vectors.

author: @christinehc

"""
import itertools
from collections import Counter
from typing import Dict, Generator, Set, Union

import numpy as np
from numpy.typing import NDArray
from ._version import __version__
from .alphabet import FULL_ALPHABETS, get_alphabet, get_alphabet_keys
from .utils import check_list


# store kmer basis set and transform new vectors into fitted basis
class KmerBasis:
    """Store kmer basis set and perform basis set transforms.

    Attributes
    ----------
    basis : list
        Description of attribute `basis`.
    basis_order : type
        Description of attribute `basis_order`.

    """

    def __init__(self):
        self.basis = []
        self.basis_order = {}

    def set_basis(self, basis):
        """Specify kmer basis set.

        Parameters
        ----------
        basis : list or array-like of str
            Ordered array of kmer strings.

        Returns
        -------
        self: object
            Fitted KmerBasis object.

        """
        if not check_list(basis):
            raise TypeError("`basis` input must be list or array-like.")

        self.basis = basis
        self.basis_order = {i: k for i, k in enumerate(basis)}

    def transform(self, vector, vector_basis):
        """Apply basis to new vector with separate basis.

        e.g. If the basis is set to a list of p kmers, the input
            vector array of size (m, n) -> (m, p).

        Note: Order is preserved from kmer arrays. Kmer count
        vectors are assumed to follow the order from corresponding
        kmer basis sets.

        Parameters
        ----------
        vector : list or array-like
            Array of size (m, n).
            Contains m vectors built from n kmers in the basis.
        vector_basis : list or array-like of str
            Array of size (n,) of ordered kmers.

        Returns
        -------
        list or array-like
            Transformed array of size (m, p).

        """
        if not check_list(vector_basis):
            raise TypeError("`vector_basis` input must be list or array-like.")

        if not isinstance(vector, np.ndarray):
            vector = np.asarray(vector)

        # make sure input vector matches shape of vector basis
        try:
            vector_size = vector.shape[1]
        except IndexError:
            vector_size = len(vector)

        if vector_size != len(vector_basis):
            raise ValueError(
                "Vector and supplied basis shapes"
                " must match (vector shape ="
                f" {vector.shape}"
                " and len(vector_basis) ="
                f" {len(vector_basis)})."
            )

        # get index order of kmers in the vector basis set
        vector_basis_order = {
            k: i if k in self.basis else np.nan for i, k in enumerate(vector_basis)
        }

        # convert vector basis into set basis
        # we'll add a dummy column for all those that
        # aren't present in the input basis
        unseen = [0] * vector.shape[0]
        vector = np.insert(vector, vector.shape[1], unseen, axis=1)

        i_convert = list()
        for i in range(len(self.basis)):
            kmer = self.basis_order[i]  # get basis set kmer in correct order
            if kmer in vector_basis_order:
                idx = vector_basis_order[kmer]  # locate kmer in the new vector
            else:
                idx = vector.shape[1] - 1
            i_convert.append(idx)

        return vector[:, i_convert]

        # correctly index into ND or 1D array
        try:
            return vector[:, i_convert]
        except IndexError:
            return vector[i_convert]


# generate all possible kmer combinations
def _generate(alphabet: Set[str], k: int):
    for c in itertools.product(alphabet, repeat=k):
        yield "".join(c)


# iterator object for kmer basis set given alphabet and k
class KmerSet:
    """Given alphabet and k, creates iterator for kmer basis set."""

    def __init__(self, alphabet: Union[str, int], k: int, kmers: list = None):
        """Initialize KmerSet object

        Parameters
        ----------
        alphabet : Union[str, int]
            Alphabet name or identifier
        k : int
            Kmer length
        kmers : list, optional
            List of kmers to manually specify basis set, by default None

        """
        self.alphabet = alphabet
        self.k = k

        # MAJOR ISSUE: with many k/alphabet combinations this
        #              will require enormous amounts of memory
        #              to be allocated - for an empty set.
        #              Need to fix this so that we don't run
        #              into this problem. Only populate those
        #              kmers that exist. I think this was crashing
        #              my computer with a k=8 alphabet=5

        if kmers is None:
            self._kmerlist = list(_generate(get_alphabet_keys(alphabet), k))
        else:
            self._kmerlist = kmers

    @property
    def kmers(self):
        return iter(self._kmerlist)


# manually reduce alphabet
def reduce(
    sequence: str, alphabet: Union[str, int], mapping: dict = FULL_ALPHABETS
) -> str:
    """Reduce sequence into character space of specified alphabet.

    Parameters
    ----------
    sequence : str
        Input sequence.
    alphabet : Union[str, int]
        Alphabet name or number (see `snekmer.alphabet`).
    mapping : dict
        Defined mapping for alphabet (the default is FULL_ALPHABETS).

    Returns
    -------
    str
        Transformed sequence.

    """
    sequence = str(sequence).rstrip("*")
    alphabet_map: Dict[str, str] = get_alphabet(alphabet, mapping=mapping)
    return sequence.translate(sequence.maketrans(alphabet_map))


# memfix: this is to reformat a list of kmers coming from multiple sequences
#         into a regular array and return that array and a list of the kmer
#         order -
def make_feature_matrix(vecs, min_filter=1, max_filter=1):
    # the input vecs is a ragged array - a list of lists of
    #     kmers that occur in each sequence (different lengths)
    kmerlist = list()
    for this in vecs:
        kmerlist.extend(this)
    kmerlist, kmercounts = np.unique(kmerlist, return_counts=True)

    # filter based on counts
    kmerlist = kmerlist[kmercounts > min_filter]

    nk = len(kmerlist)
    # result = np.zeros(len(kmerlist)*len(vecs)).reshape(len(vecs),len(kmerlist))

    result = list()
    for i in range(len(vecs)):
        this = np.zeros(nk)
        this[np.isin(kmerlist, vecs[i])] = 1
        result.append(this)

    return result, kmerlist


class KmerVec:
    def __init__(self, alphabet: Union[str, int], k: int):
        self.alphabet = alphabet
        self.k = k
        self.char_set = get_alphabet_keys(alphabet)
        self.vector = None
        self.basis = KmerBasis()
        self.snekmer_version = __version__
        # self.kmer_set = KmerSet(alphabet, k)

    def set_kmer_set(self, kmer_set=list()):
        self.kmer_set = KmerSet(self.alphabet, self.k, kmer_set)
        self.basis.set_basis(kmer_set)

    # iteratively get all kmers in a string
    def _kmer_gen(self, sequence: str) -> Generator[str, None, None]:
        """Generator object for segment in string format"""
        i = 0
        n = len(sequence) - self.k + 1

        # iterate thru sequence in blocks of length k
        while i < n:
            kmer = sequence[i : i + self.k]
            if set(kmer) <= self.char_set:
                yield kmer
            i += 1

    # not used: iterate using range() vs. while loop
    @staticmethod
    def _kmer_gen_str(sequence: str, k: int) -> Generator[str, None, None]:
        """Generator object for segment in string format"""
        for n in range(0, len(sequence) - k + 1):
            yield sequence[n : n + k]

    # generate kmer vectors with bag-of-words approach
    def vectorize(self, sequence: str) -> NDArray:
        """Transform sequence into representative kmer vector.

        Parameters
        ----------
        sequence : str
            Input sequence.

        Returns
        -------
        NDArray
            Vector representation of sequence as kmer counts vector.

        """
        N = len(self.char_set) ** self.k

        kmers = list(self._kmer_gen(sequence))
        kmer2count = Counter(kmers)

        # Convert to vector of counts
        # vector = np.zeros(N)

        # memfix change
        vector = {}

        for i, word in enumerate(self.kmer_set.kmers):
            vector[i] += kmer2count[word]

        # Convert to frequencies
        # vector /= sum(kmer2count.values())

        return vector

    def reduce_vectorize(self, sequence: str) -> NDArray:
        """Simplify and vectorize sequence into reduced kmer vector.

        Parameters
        ----------
        sequence : str
            Input sequence.

        Returns
        -------
        NDArray
            Vector representation of sequence as reduced kmer vector.

        """
        # N = len(self.char_set) ** self.k

        reduced = reduce(sequence, alphabet=self.alphabet, mapping=FULL_ALPHABETS)
        kmers = list(self._kmer_gen(reduced))
        # kmer2count = Counter(kmers)

        vector = np.array(kmers, dtype=str)

        # memfix change
        # this changes the output from a list to a dict
        # vector = {}
        # for kmer in kmers:
        #    vector[kmer] = 1

        # Convert to vector of counts
        # vector = np.zeros(N)
        # for i, word in enumerate(self.kmer_set.kmers):
        #    vector[i] += kmer2count[word]

        # Convert to frequencies
        # vector /= sum(kmer2count.values())

        return vector

    def harmonize(self, record, kmerlist):
        """_summary_

        Parameters
        ----------
        record : _type_
            _description_
        kmerlist : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        return self.basis.transform(record, kmerlist)



from . import alphabet
from . import utils
from . import io
from . import plot
from . import score
from . import model
from . import cluster
from . import vectorize
from . import report
from . import _version

# from . import walk

__version__ = _version.__version__



# ---------------------------------------------------------
# Imports
# ---------------------------------------------------------
import pickle
from datetime import datetime

import pandas as pd
import snekmer as skm
from sklearn.model_selection import train_test_split, StratifiedKFold

# ---------------------------------------------------------
# Files and Parameters
# ---------------------------------------------------------

config = snakemake.config

# ---------------------------------------------------------
# Run script
# ---------------------------------------------------------

# log script start time
start_time = datetime.now()
label = (
    config["score"]["lname"] if str(config["score"]["lname"]) != "None" else "label"
)  # e.g. "family"
with open(snakemake.log[0], "a") as f:
    f.write(f"start time:\t{start_time}\n")

# get kmers for this particular set of sequences
kmers = skm.io.load_pickle(snakemake.input.kmerobj)

# tabulate vectorized seq data
data = list()
kmer_sets = list()
for f in snakemake.input.data:
    kmerlist, df = skm.io.load_npz(f)
    data.append(df)
    kmer_sets.append(kmerlist[0])

# loading all other models and harmonize basis sets
for i in range(len(data)):
    df = data[i]
    kmerlist = kmer_sets[i]
    vecs = skm.utils.to_feature_matrix(df["sequence_vector"].values)
    df["sequence_vector"] = kmers.harmonize(vecs, kmerlist).tolist()

data = pd.concat(data, ignore_index=True)
data["background"] = [f in snakemake.input.bg for f in data["filename"]]

# log conversion step runtime
skm.utils.log_runtime(snakemake.log[0], start_time, step="files_to_df")

# parse family names and only add if some are valid
families = [
    skm.utils.get_family(
        skm.utils.split_file_ext(fn)[0], regex=config["input_file_regex"]
    )
    for fn in data["filename"]
]
if any(families):
    data[label] = families

# binary T/F for classification into family
family = skm.utils.get_family(snakemake.wildcards.nb)
binary_labels = [True if value == family else False for value in data[label]]

# define k-fold split indices
if config["model"]["cv"] > 1:
    cv = StratifiedKFold(n_splits=config["model"]["cv"], shuffle=True)

    # stratify splits by [0,1] family assignment
    for n, (i_train, _) in enumerate(cv.split(data["sequence_vector"], binary_labels)):
        data[f"train_cv-{n + 1:02d}"] = [idx in i_train for idx in data.index]

elif config["model"]["cv"] in [0, 1]:
    i_train, _ = train_test_split(data.index, stratify=binary_labels)
    data["train"] = [idx in i_train for idx in data.index]

# generate family scores and object
scorer = skm.score.KmerScorer()
scorer.fit(
    list(kmers.kmer_set.kmers),
    data,
    skm.utils.get_family(snakemake.wildcards.nb, regex=config["input_file_regex"]),
    label_col=label,
    vec_col="sequence_vector",
    **config["score"]["scaler_kwargs"],
)

# append scored sequences to dataframe
data = data.merge(
    pd.DataFrame(scorer.scores["sample"]), left_index=True, right_index=True
)
if data.empty:
    raise ValueError("Blank df")

# save score loadings
class_probabilities = (
    pd.DataFrame(scorer.probabilities, index=scorer.kmers.basis)
    .reset_index()
    .rename(columns={"index": "kmer"})
)

# log time to compute class probabilities
skm.utils.log_runtime(snakemake.log[0], start_time, step="class_probabilities")

# this will be redundant with the output csv file - except that
#      it will have the harmonized matrix included with it
with open(snakemake.output.matrix, "wb") as f:
    pickle.dump(data, f)

# save all files to respective outputs
delete_cols = ["vec", "sequence_vector"]
for col in delete_cols:
    if col in class_probabilities.columns:
        class_probabilities = class_probabilities.drop(columns=col)
data.drop(columns="sequence_vector").to_csv(
    snakemake.output.data, index=False, compression="gzip"
)

class_probabilities.to_csv(snakemake.output.weights, index=False, compression="gzip")
with open(snakemake.output.scorer, "wb") as f:
    pickle.dump(scorer, f)

# record script endtime
skm.utils.log_runtime(snakemake.log[0], start_time)



"""alphabet: Built-in alphabet mappings for Snekmer.

2021.08.11 - Alphabet names have been reworked per @wichne
             (@christinehc)

author(s): @biodataganache, @wichne

"""
# imports
from typing import Dict, Mapping, Set, Union

# define standard amino acid alphabet
StandardAlphabet = "AILMVFYWSTQNCHDEKRGP"
AA_SELF_MAPPING = {a: a for a in StandardAlphabet}

# post-translational modification
PTM_CHARS = "-_!^#$@.%&"
PTM_SELF_MAPPING = {c: c for c in PTM_CHARS}

# define alphabet names and ordering
ALPHABET_ORDER = {
    0: "hydro",
    1: "standard",
    2: "solvacc",
    3: "hydrocharge",
    4: "hydrostruct",
    5: "miqs",
}

# define alphabets (from Utils.SIEVEInit)
ALPHABETS = {
    # 2-value hydrophobicity alphabet taken
    # from Arnold, et al. 2009. PLoS Pathogens 5(4): e1000376
    "hydro": {"SFTNKYEQCWPHDR": "S", "VMLAIG": "V", "_keys": "SV"},
    # 'standard' reduction alphabet taken
    # from Arnold, et al. 2009. PLoS Pathogens 5(4): e1000376
    "standard": {
        "AGILMV": "A",  # hydrophobic
        "PH": "P",  # hydrophilic
        "FWY": "F",  # aromatic
        "NQST": "N",  # polar
        "DE": "D",  # acidic
        "KR": "K",  # alkaline
        "C": "C",  # ionizable
        "_keys": "APFNDKC",
    },
    # Solvent accessibility alphabet
    # from Bacardit, et al. 2009. BMC Bioinformatics 10:6
    "solvacc": {"CILMVFWY": "C", "AGHST": "A", "PDEKNQR": "P", "_keys": "CAP"},
    # 2-value hydrophobicity with charged residues as a third
    # category. Made by @biodataganache.
    "hydrocharge": {
        "SFTNYQCWPH": "L",  # hydrophilic (L-ove)
        "VMLAIG": "H",  # hydrophobic (H-ate)
        "KNDR": "C",  # charged (C-harged)
        "_keys": "LHC",
    },
    # 2-value hydrophobicity with structural-breakers as a third category
    # Made by @biodataganache
    "hydrostruct": {"SFTNKYEQCWHDR": "L", "VMLAI": "H", "PG": "B", "_keys": "LHB"},
    # MIQS alphabet
    # Made by @wichne
    "miqs": {
        "A": "A",  # Alanine
        "C": "C",  # Cysteine
        "DEN": "D",  # acidicish
        "FWY": "F",  # aromatic
        "G": "G",  # glycine
        "H": "H",  # histidine
        "ILMQV": "I",  # hydrophobicish
        "KR": "K",  # alkaline
        "P": "P",  # proline
        "ST": "S",  # hydroxyl
        "_keys": "ACDFGHIKPS",
    },
    # # identity
    # "None": {**AA_SELF_MAPPING, "_keys": ALL_AMINO_ACIDS}  # OU
    # post-translational modification alphabet
    "ptm": {
        **AA_SELF_MAPPING,
        **PTM_SELF_MAPPING,
        "_keys": StandardAlphabet + PTM_CHARS,
    },
    "None": AA_SELF_MAPPING,
}

# reconfigure alphabet dict into "long-form"
FULL_ALPHABETS: Dict[str, dict] = {a: {} for a in ALPHABETS.keys()}
for alphabet, mapping in ALPHABETS.items():
    for k, v in mapping.items():
        if k == "_keys":
            continue
        elif len(k) > 1:
            FULL_ALPHABETS[alphabet].update({k[i]: v for i in range(len(k))})
        else:
            FULL_ALPHABETS[alphabet].update({k: v})

# create generic alphabet identifiers
ALPHABET_ID = {
    f"RED{n}": {v: k for k, v in ALPHABETS[ALPHABET_ORDER[n]].items()}
    for n in range(len(ALPHABET_ORDER))
}

# alphabet to alphabet code
ALPHABET2ID = {ALPHABET_ORDER[n]: f"RED{n}" for n in range(len(ALPHABET_ORDER))}


# SIEVEInit alphabet grabbing function
def get_alphabets():
    """Return all alphabet mappings.

    Returns
    -------
    dict of dicts
        {'alphabet': {'key': 'mapping', ...}} for each alphabet

    """
    return ALPHABETS


def check_valid(alphabet: Union[str, int]) -> None:
    """Check validity of input alphabet vs. defined list.

    Parameters
    ----------
    alphabet : Union[str, int]
        Alphabet name or identifier.

    Raises
    ------
    ValueError
        If `alphabet` does not match a pre-defined alphabet.

    """
    if (
        # raise error if alphabet not included in pre-defined set
        (
            # doesn't match integer designations
            (alphabet not in range(len(ALPHABETS)))
            # doesn't match str name designations
            and (alphabet not in ALPHABETS)
        )
        # or doesn't match None (no mapping)
        and (str(alphabet) != "None")
    ):  # and config['alphabet'] != 'custom':
        raise ValueError(
            "Invalid alphabet specified; alphabet must be a"
            " string (see snekmer.alphabet) or integer"
            " n between"
            f" {min(list(ALPHABET_ORDER.keys()))}"
            " and"
            f" {max(list(ALPHABET_ORDER.keys()))}"
            "."
        )
    return


def get_alphabet(
    alphabet: Union[str, int], mapping: dict = ALPHABETS
) -> Dict[str, str]:
    """Short summary.

    Parameters
    ----------
    alphabet : Union[str, int]
        Alphabet name (as str) or alphabet id (as int).
        Must be one of the following:
            0: "hydro",
            1: "standard",
            2: "solvacc",
            3: "hydrocharge",
            4: "hydrostruct",
            5: "miqs"
        or None.
    mapping : dict
        All alphabet maps (the default is ALPHABETS).

    Returns
    -------
    dict
        Dictionary map of amino acids to alphabet character.

    Raises
    ------
    ValueError
        Raised if alphabet not in pre-defined list.

    """
    check_valid(alphabet)

    # enforce string for dict key
    if alphabet is None:
        alphabet = str(alphabet)

    if isinstance(alphabet, int):
        alphabet = ALPHABET_ORDER[alphabet]
    return mapping[alphabet]


def get_alphabet_name(
    alphabet: Union[str, int], mapping: dict = ALPHABETS
) -> Dict[str, str]:
    """Get alphabet name given any input type.

    Parameters
    ----------
    alphabet : Union[str, int]
        Alphabet name (as str) or alphabet id (as int).
        Must be one of the follwing:
            0: "hydro",
            1: "standard",
            2: "solvacc",
            3: "hydrocharge",
            4: "hydrostruct",
            5: "miqs"
    mapping : dict
        All alphabet maps (the default is ALPHABETS).

    Returns
    -------
    dict
        Dictionary map of amino acids to alphabet character.

    Raises
    ------
    ValueError
        Raised if alphabet not in pre-defined list.

    """
    check_valid(alphabet)

    # enforce string for dict key
    if alphabet is None:
        alphabet = str(alphabet)

    if isinstance(alphabet, int):
        alphabet = ALPHABET_ORDER[alphabet]
    return alphabet


def get_alphabet_keys(
    alphabet: Union[str, int], mapping: Dict[str, dict] = FULL_ALPHABETS
) -> Set[str]:
    """Retrieve keys for specified alphabet.

    Parameters
    ----------
    alphabet : Union[str, int]
        Description of parameter `alphabet`.
    mapping : Mapping[dict]
        Description of parameter `mapping` (the default is FULL_ALPHABETS).

    Returns
    -------
    dict
        Description of returned object.

    """
    # enforce string for dict key
    if alphabet is None:
        alphabet = str(alphabet)

    alphabet_map = get_alphabet(alphabet, mapping)
    if "_keys" in alphabet_map.keys():
        alphabet_map.pop("_keys")
    return set(alphabet_map.values())


# def add_alphabet(alphabet_name, mapping):
#     return



"""utils: Snekmer utility functions.

author: @christinehc

"""
# imports
import collections.abc
import datetime
import re

from os.path import basename, splitext
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray


# functions
def _format_timedelta(timedelta: datetime.timedelta) -> str:
    """Format datetime.timedelta object as str with units.

    Parameters
    ----------
    timedelta : datetime.timedelta object
        A datetime.timedelta object.

    Returns
    -------
    str
        Formatted string with the following format:
            %%h %%m %%.%%s
            (e.g. '01h 41m 20.01s')

    """
    # get milliseconds and round
    milliseconds = round(timedelta.microseconds / 1000)

    # get H, M, S
    minutes, seconds = divmod(timedelta.seconds, 60)
    hours, minutes = divmod(minutes, 60)

    return f"{hours}h {minutes:02}m {seconds:02}.{milliseconds}s"


def log_runtime(
    filename: str, start_time: datetime.datetime, step: Optional[str] = None
) -> None:
    """Append runtimes to specified log file.

    Parameters
    ----------
    filename : str
        /path/to/log/file.
    start_time : datetime.datetime object
        Previous `datetime.now()` result.
    step : str or None (default: None)
        Optional name for step

    Returns
    -------
    None
        Appends runtime information to file and returns no output.

    """
    end_time = datetime.datetime.now()
    if step is None:
        step = "end"

    with open(filename, "a") as f:
        f.write(f"start time:\t{start_time}\n")
        f.write(f"{step} time:\t{end_time}\n")
        f.write(f"total time:\t{_format_timedelta(end_time - start_time)}")
    return


def check_list(array: Any) -> bool:
    """Check whether input is a sequence, list, or array-like object.

    Parameters
    ----------
    array : type
        Input array (or any parameter type).

    Returns
    -------
    bool
        Returns True if input is a list/sequence, array, or Series.

    """
    if not isinstance(array, (collections.abc.Sequence, np.ndarray, pd.Series)):
        return False
    return True


def get_family(
    filename: str, regex: str = r"[a-z]{3}[A-Z]{1}", return_first: bool = True
) -> Union[str, List[str]]:
    """Extract family from filename given regular expression format.

    Note: If no family matching the given regular expression is
    found, the original filename is returned, stripped of any file
    extension (i.e. the file's basename).

    Parameters
    ----------
    filename : str
        path/to/filename.ext
    regex : str or r-string or None
        Regular expression for matching a family name
        (default: "[a-z]{3}[A-Z]{1}").
        The default expression is 3 lowercase letters followed
            by one uppercase letter. To write a custom regular
            expression, see https://docs.python.org/3/library/re.html
            for more details on using the built-in re library.
        If None, returns the full file basename.
    return_first : bool
        If True, returns only the first occurrence (default: True).
        If False, returns a list of all occurring regex matches.

    Returns
    -------
    str or list
        Family name or names


    """
    # extract and simplify file basename
    filename = basename(filename)

    # account for directories
    if "." not in filename:  # and filename[-1] == "/"
        filename = f"{filename}.dir"
    s = "_".join(filename.split(".")[:-1]).replace("-", "_").replace(" ", "_")

    # explicitly define regex as an r-string
    if regex is not None:
        regex = r"{}".format(regex)
    else:
        return filename
    search = re.search(regex, s)

    # return list output
    if not return_first:
        re.findall(regex, s)

    # return string output
    if search is not None:
        return search.group()
    return filename.split(".")[0]


def split_file_ext(filename: str) -> Tuple[str, str]:
    """Split file.ext into (file, ext).

    Ignores ".gz" for gzipped files; e.g. "file.ext.gz" returns
    ("file", "ext") rather than ("file.ext", "gz").

    Parameters
    ----------
    filename : str
        /path/to/file.ext.

    Returns
    -------
    (str, str)
        Tuple containing (file, ext) for file.ext.

    """
    filename = basename(filename)

    # for compressed files, returns (filename, ext) without .gz
    if splitext(filename)[1] == ".gz":
        return (
            splitext(splitext(filename)[0])[0],
            splitext(splitext(filename)[0])[1].lstrip("."),
        )

    # otherwise, returns (filename, ext)
    return splitext(filename)[0], splitext(filename)[1].lstrip(".")


def to_feature_matrix(
    array: List, length_array: Optional[Union[List, NDArray, pd.Series]] = None
) -> NDArray:
    """Create properly shaped feature matrix for kmer scoring.

    Parameters
    ----------
    array : numpy.ndarray or list or array-like
        2-D array-like collection of kmer vectors.
        The assumed format is rows = sequences, cols = kmers.

    Returns
    -------
    numpy.ndarray of numpy.ndarrays
        2D array version of the 2D array-like input.

    """
    if length_array is None:
        length_array = np.ones(len(array))
    array = [np.array(a) / length for a, length in zip(array, length_array)]
    return np.asarray(array)


def count_n_seqs(filename: str) -> int:
    """Count number of sequences in a file.

    Parameters
    ----------
    filename : str
        /path/to/sequence/file.fasta
        Other text formatted sequence files (.faa, etc.) also work.

    Returns
    -------
    int
        Number of sequences contained within the input file.

    """
    return len([1 for line in open(filename) if line.startswith(">")])


def check_n_seqs(filename: str, k: int, show_warning: bool = True) -> bool:
    """Check that a file contains at least k sequences.

    Parameters
    ----------
    filename : str
        /path/to/sequence/file.fasta
        Other text formatted sequence files (.faa, etc.) also work.
    k : int
        Minimum threshold.
    show_warning : bool, optional
        When True, if len(file) < k, a warning is displayed;
        by default True.

    Returns
    -------
    bool
        True if len(file) < k; False otherwise.

    """
    n_seqs = len([1 for line in open(filename) if line.startswith(">")])
    if (n_seqs < k) and (show_warning):
        print(
            f"\nWARNING: {filename} contains an insufficient"
            " number of sequences for model cross-validation"
            " and will thus be excluded from Snekmer modeling."
            f" ({k} folds specified in config; {n_seqs}"
            " sequence(s) detected.)\n"
        )
    return n_seqs >= k



"""score: Kmer-based scoring and similarity analysis.

author: @christinehc, @biodataganache

"""
# imports
import numpy as np
import pandas as pd

from ._version import __version__
from .vectorize import KmerBasis
from .utils import to_feature_matrix
from sklearn.metrics.pairwise import pairwise_distances


# scaling object for reducing kmers used in scoring
class KmerScoreScaler:
    """Scale and reduce kmer set based on scores.

    Parameters
    ----------
    scores : list or numpy.ndarray
        Description of parameter `model`.
    n : int
        Feature cutoff (top n kmer features to include)

    Attributes
    ----------
    n : int or float
        Feature cutoff (top n kmer features).
        Accepts either an integer of the top n features, or a float
        indicating a percentage (n = 0.1 for the top 10% of features)
    input_shape : tuple
        Input array shape.
    kmer_basis_idx : numpy.ndarray
        List of indices from kmer feature vector making up the
        specified basis set, ordered from most to least important.
    kmer_basis_score : list of tuples
        List of tuples for kmer feature vector making up the
        specified basis set, ordered from most to least important.
        Tuples are: (index, score) for each listed feature.
    scores : numpy.ndarray
        Array of scores for kmer features.

    """

    def __init__(self, n=100):
        """Initialize KmerScaler object.

        """
        self.n = n
        self.input_shape = None
        self.scores = None
        self.basis_index = None
        self.basis_score = dict()

    def fit(self, scores):
        """Fit scaler based on list of scores.

        Parameters
        ----------
        scores : list or numpy.ndarray
            Array of scores for kmer features.
        threshold : float
            Numerical cutoff for kmer feature scores.

        Returns
        -------
        None
            Fits KmerScoreScaler() object.

        """
        if not isinstance(scores, np.ndarray):
            scores = np.array(scores)
        self.scores = scores
        self.input_shape = scores.shape

        # option 1: set score threshold as limit -- REMOVED
        # if threshold is not None:
        #     indices = np.where(np.array(scores > threshold))[0]

        if self.n is not None:
            # option 2: set # of features as limit
            if (isinstance(self.n, int)) and (self.n >= 1):
                indices = scores.ravel().argsort()[: -self.n - 1 : -1]
            # option 3: set % of features as limit
            elif (isinstance(self.n, float)) and (self.n < 1):
                n = int(np.floor(self.n * len(scores)))
                indices = scores.ravel().argsort()[: -n - 1 : -1]
            else:
                raise ValueError(
                    "Invalid input format for `n`"
                    " (must be either int > 1, or"
                    " 0.0 < float < 1.0)."
                )
        else:
            raise ValueError("One of either `threshold` or `n` must" " be specified.")

        # store basis set and indices as attributes
        self.basis_index = indices
        self.basis_score = [
            (i, score) for i, score in zip(self.basis_index, scores[self.basis_index])
        ]
        return

    def transform(self, array):
        """Reduce vector to kmer basis set.

        Parameters
        ----------
        array : list or numpy.ndarray
            Unscaled kmer vector or matrix.

        Returns
        -------
        numpy.ndarray
            Scaled kmer vector or matrix.

        """
        if self.basis_index is None or not any(self.basis_index):
            raise AttributeError(
                "Kmer basis set not defined;" " must fit scores to scaler."
            )
        # if not isinstance(array, np.ndarray):
        #     array = np.array(array)
        if isinstance(array[0], list):
            array = to_feature_matrix(array)

        # input vec must match original input shape for indexing
        if len(array.shape) < 2:  # account for 1D arrays
            array_size = len(array)
            input_size = self.input_shape[0]
            if array_size != input_size:
                raise ValueError(
                    f"Input vector shape {array.shape}"
                    f" does not match fitted vector shape"
                    f" {self.input_shape}."
                )
            return array[self.basis_index]

        # ND arrays
        array_size = array.shape
        input_size = self.input_shape
        return array[:, self.basis_index]


# functions
def connection_matrix_from_features(feature_matrix, metric="jaccard"):
    """Calculate similarities based on features output from main().

    Parameters
    ----------
    feature_matrix : numpy.ndarray
        Feature matrix in the form of rows (sequences), columns
        (kmers), where values are kmer counts for each protein.
    metric : str
        Metric used for pairwise distance computation (see
            sklearn.metrics.pairwise documentation).

    Returns
    -------
    numpy.ndarray
        Square matrix with the similarity scores of pairwise
        relationships between proteins.

    """
    if metric == "jaccard":
        sim = 1 - pairwise_distances(feature_matrix, metric="hamming")  # .T?
    else:
        sim = pairwise_distances(feature_matrix, metric=metric)
    return sim


def _apply_probability(kmer, label, compare_label):
    """Compute binary probability of a label for a kmer.

    Parameters
    ----------
    kmer : int
        Kmer count (assumes binary presence/absence (0, 1)).
    label : str
        Label value for a given kmer.
    compare_label : str
        Label value to assess a given kmer against.

    Returns
    -------
    list
        Binary probability (0, 1) of valid kmer with given label
        for all kmers in the kmer count array.
            e.g. [0, 1, 1, 0] for an input kmer array of length 4

    """
    return (kmer == 1) * 1 * (label == compare_label)


def _binarize(feature_matrix):
    """Convert feature counts into binary presence/absence.

    Parameters
    ----------
    feature_matrix : array (list, numpy.ndarray, pandas.Series)
        Feature matrix, where each row represents a kmer and each
        column represents a sequence

    Returns
    -------
    numpy.ndarray
        Description of returned object.

    """
    return (feature_matrix > 0) * 1.0


def _get_kmer_presence(feature_matrix, label_match):
    """Compute kmer presence of a given label in sequences.

    Parameters
    ----------
    feature_matrix : array (list, numpy.ndarray, pandas.Series)
        Feature matrix of shape (n, m), where each row represents
        a kmer and each column represents a sequence.
        In other words, m must equal len(labels).
    label_match : list or numpy.ndarray or pandas.Series
        Array of shape (m, ) describing label matches.
        e.g. contains 1 when kmer matches a label, and 0 otherwise

    Returns
    -------
    numpy.ndarray
        Array describing kmer presence of a given label

    """
    # convert feature count matrix into binary presence/absence
    feature_matrix = _binarize(feature_matrix)

    # for each row in feature matrix, compute each value's probability
    # no multiprocessing-- using numpy native parallelization
    result = list()
    for kmer in feature_matrix:
        kmer_match = np.equal(kmer, 1)  # array of presence matches
        ones = np.ones(len(kmer))
        match = np.multiply(np.multiply(kmer_match, label_match), ones)
        result.append(np.sum(match))
    return result


def _score(p_pos, p_neg, w=1.0):
    """Score probability vs. weighted negative probability.

    e.g. P(pos) - (w * P(neg))

    Parameters
    ----------
    p_pos : float
        Probability of positive assignment.
    p_neg : float
        Probability of negative assignment.
    w : float (default: 1.0)
        Weight for negative assignment probability.

    Returns
    -------
    float
        Total probability of positive identification, minus
        contribution to negative identification.

    """
    # p_in = results[in_label][col]
    # p_out = np.sum([results[fam]['probability'] for fam in out_labels], axis=0)
    # p_bg = bg_labels
    return p_pos - (w * p_neg)  # - (w_bg * p_bg)


def score(results, in_label, out_labels, w=1.0, col="probability", neg_only=False):
    """Score kmer from kmer family probabilities.

    The scoring method used is as follows:
        Score = P(in family) - (w_out * P(out of family)) - P(bg)

    Parameters
    ----------
    results : dict or pandas.DataFrame
        Prior results containing all kmer family probabilities.
        Requires nested dict or multi-index, e.g.:
            results[family][`col`] = pos kmer family probability
    in_label : str
        Name of "in-family" label.
    out_labels : list or array-like
        Array of "out-of-family" labels.
    w : float (default: 1.0)
        Multiplier for out-of-family probability subtraction.
    col : str (default: 'probability')
        Name of column containing kmer probabilities.
    neg_only : bool (default: False)
        If True, only returns the negative probability (-w * P_out).

    Returns
    -------
    float
        Score for in-family kmer assignment.

    """
    p_in = results[in_label][col]
    p_out = np.sum([results[fam]["probability"] for fam in out_labels], axis=0)

    # negative probability only
    if neg_only:
        return _score(0, p_out, w=w)

    return _score(p_in, p_out, w=w)


def _parse_score_method(method, bg=None, **kwargs):
    """Translate method name into corresponding score method(s).

    Parameters
    ----------
    method : str
        Method name:
            'default' - Default score method
            'bg_only' - Score sequences as though they are all background
            'both' - Compute both score types
    bg : type
        Description of parameter `bg`.
    **kwargs : type
        Description of parameter `**kwargs`.

    Returns
    -------
    type
        Description of returned object.

    """
    # restrict method options
    if method not in ["default", "bg_only", "both"]:
        raise ValueError(
            "Parameter `method` must be one of the"
            " following: 'default', 'bg_only', 'both'."
        )

    # check that background is provided when required
    if (method in ["bg_only", "both"]) and (bg is None):
        raise ValueError(
            "Background sequence vectors must be"
            " provided if `method`='bg_only' or"
            " 'both'."
        )

    # define score methods
    method2name = {
        "default": "default_score",
        "bg_only": "bg_subtracted_score",
        "both": ["default_score", "bg_subtracted_score", "combined_score"],
    }

    # parse scoring methods
    if method == "both":
        return "combined_score", score
        # compute score based on (label, not label) assignment
        # weight = 1 / (len(labels) - 1)
        # for l in labels:
        #     o = labels[labels != l]  # other labels
        #     results[l]['score'] = score(results, l, o, w=weight)
    return method2name[method], score


def feature_class_probabilities(
    feature_matrix, labels, kmers=None, bg_matrix=None, method="default"
):
    """Calculate probabilities for features being in a defined class.

    Note: only coded to work for the binary case (2 classes).

    Parameters
    ----------
    feature_matrix : array (list, numpy.ndarray, pandas.Series)
        Feature matrix of shape (n, m), where each row represents
        a kmer and each column represents a sequence.
        In other words, m must equal len(labels).
    labels : list or numpy.ndarray or pandas.Series
        Class labels of shape (m, ) describing feature matrix.
        Must have as many entries as the number of feature columns.
    kmers : list or None (default: None)
        Optional kmer identifiers mapping to kmer columns (shape (n,)).
    method : str (default: "default")
        Specify scoring method:
            "default" : Score in-family vs. out-of-family assignment
                P(family) - (weight * P(out of family))
            "bg_only" : Score in-family vs. background
                P(family) - P(background)
            "both" : Score in-family vs. out-of-family and background
                P(family) - P(background) - (weight * P(out of family))
    bg_matrix : array or None (default: None)
        Feature matrix of shape (n, o), where each row represents
        a kmer and each column represents a background sequence.

    Returns
    -------
    numpy.ndarray (df=False) or pandas.DataFrame (df=True)
        Array or DataFrame containing scoring and probability
        results for each family.

    """
    # ensure labels are given as a numpy array
    if isinstance(labels, (np.ndarray, list, pd.Series)):
        labels = np.array(labels, dtype=object)
    else:
        raise TypeError("Labels must be list- or array-like.")

    # parse score function
    method_name, score_function = _parse_score_method(method)

    # kmer labels
    if kmers is None:
        kmers = np.array([n for n in range(len(feature_matrix))])
    if len(kmers) != len(feature_matrix):
        # print(len(kmers), len(feature_matrix), feature_matrix.shape)
        raise ValueError("Kmer array shape is mismatched.")

    # check that labels are the same size as number of examples
    if len(feature_matrix.T) != len(labels):
        raise ValueError("Input shapes are mismatched.")

    # get only unique labels
    unique_labels = np.unique(labels)

    # iteratively score all feature in input matrix
    n_seqs = [len(np.hstack(np.where(labels == lname))) for lname in unique_labels]
    results = {label_name: {} for label_name in labels}
    for i, l in enumerate(unique_labels):
        label_match = np.equal(labels, l)

        # get score arrays based on label match, normalized by #seqs
        presence = _get_kmer_presence(feature_matrix, label_match)
        norms = (1 / n_seqs[i]) * np.ones(len(presence))

        # if no kmers specified, save numerical range
        results[l]["kmer"] = np.array(kmers)
        results[l]["count"] = np.asarray(presence, dtype=int)
        results[l]["probability"] = np.asarray(presence * norms, dtype=float)

    # compute score based on (label, not label) assignment
    label_weight = len(unique_labels)  # if 1 label, no weight (w=1)
    if len(unique_labels) > 1:  # if >1 label, weight = 1 / (n - 1)
        label_weight = label_weight - 1
    weight = 1 / label_weight
    for l in unique_labels:
        o = unique_labels[unique_labels != l]  # other labels
        results[l]["score"] = score(results, l, o, w=weight)

    # reformat as long-form dataframe
    results = pd.DataFrame(results).T.reset_index().rename(columns={"index": "label"})
    results = results.set_index(["label"]).apply(pd.Series.explode).reset_index()
    return results


# from Jason
# This code will apply the feature class probabilities derived in the previous functions
#   to a new set of examples as a weighted score. The idea here is to use a very simple
#   approach to classification.
def apply_feature_probabilities(feature_matrix, scores, scaler=False, **kwargs):
    """Calculate class probability scores based on kmer vectors.

    Parameters
    ----------
    feature_matrix : array (list, numpy.ndarray, pandas.Series)
        Feature matrix of shape (n, m), where each row represents
        a kmer and each column represents a sequence.
        In other words, len(feature_matrix.T) must equal len(labels)
    scores : array (list, numpy.ndarray, pandas.Series)
        Array of shape (m,) of kmer probability scores.
        Scores must be in the same kmer order as in `feature_matrix`.
    scaler : obj or bool (default: False)
        If True, performs scaling to reduce kmer features.
        If False, does not perform scaling to reduce kmer features.
        If obj is supplied, the scaling obj is applied to reduce
            kmer features.
    **kwargs : dict
        Keyword arguments for KmerScoreScaler().

    Returns
    -------
    numpy.ndarray
        Feature matrix weighted

    """
    # optionally apply scaling
    if scaler and isinstance(scaler, bool):
        scaler = KmerScoreScaler(**kwargs)
        scaler.fit(scores)
    if isinstance(scaler, KmerScoreScaler):
        # scaler.fit(scores)
        scores = scaler.transform(scores)
        feature_matrix = scaler.transform(feature_matrix)

    feature_matrix = _binarize(feature_matrix)

    score_matrix = scores * feature_matrix
    score_totals = np.array([np.sum(arr) for arr in score_matrix])

    return score_totals


class KmerScorer:
    """Score kmer vectors based on summed probability scores.

    Attributes
    ----------
    kmers : :obj:`~KmerBasis`
        Basis set object for kmer standardization.
    matrix : numpy.ndarray
        Feature matrix for all sequences.
    labels : list or array-like of str
        Array of all labels in overall dataset.
    primary_label : str
        Primary label identifier for fitted sequence data.
    i_background : dict of list
        Description of attribute `i_background`.
    i_label : dict
        Description of attribute `i_label`.
    scaler : type
        Description of attribute `scaler`.
    scores : type
        Description of attribute `scores`.
    score_norm : type
        Description of attribute `score_norm`.
    probabilities : type
        Description of attribute `probabilities`.

    """

    def __init__(self):
        self.kmers = KmerBasis()
        self.primary_label = None
        self.scaler = None
        self.scores = {"sample": {}, "background": {}}
        self.score_norm = None
        self.probabilities = {"sample": {}, "background": {}}
        self.snekmer_version = __version__

    # load list of kmers and both seq and bg feature matrices
    def fit(
        self,
        kmers,
        data,
        label,
        label_col="family",
        bg_col="background",
        vec_col="vector",
        **scaler_kwargs,
    ):
        """Short summary.

        Parameters
        ----------
        kmers : type
            Description of parameter `kmers`.
        data : type
            Description of parameter `data`.
        label : type
            Description of parameter `label`.
        label_col : type
            Description of parameter `label_col`.
        bg_col : type
            Description of parameter `bg_col`.
        **scaler_kwargs : dict
            Keyword arguments for
            :obj:`~snekmer.model.KmerScoreScaler`.

        Returns
        -------
        type
            Description of returned object.

        """
        # save primary family label
        self.primary_label = label

        # step 00: set kmer basis
        self.kmers.set_basis(kmers)

        # step 0: get indices of sample and background sequences
        i_background = {
            "sample": list(data.index[~data[bg_col]]),
            "background": list(data.index[data[bg_col]]),
        }

        # step 0: get feature matrix and all labels
        labels = data[label_col].values

        # print(data[vec_col].shape)
        # print(data[vec_col])

        x = len(data[vec_col])
        y = len(data[vec_col][0])
        # print(x,y)

        matrix = np.zeros(x * y).reshape((x, y))
        # print(matrix.shape)

        for i in range(x):
            for j in range(y):
                value = data[vec_col][i]
                value = value[j]
                matrix[i, j] = value
        # matrix = np.asarray(np.concatenate(data[vec_col])).reshape((len(data[vec_col]), len(data[vec_col][0])))
        # print(matrix.shape)

        # step 0: get indices for label (family) ids
        # i_label = {
        #     ll: list(data.index[data[label_col] == ll])
        #     for ll in data[label_col].unique()
        # }

        # step 1: score sample sequences and fit score scaler
        sample_matrix = matrix[i_background["sample"]]
        s_labels = labels[i_background["sample"]]

        probas = feature_class_probabilities(
            sample_matrix.T, s_labels, kmers=self.kmers.basis
        )
        self.probabilities["sample"] = probas[probas["label"] == self.primary_label][
            "score"
        ].values

        # step 2: fit scaler to the sample data (ignore the background)
        self.scaler = KmerScoreScaler(**scaler_kwargs)
        self.scaler.fit(
            probas[probas["label"] == label]["probability"]
        )  # probas is an ordered list of scores from kmers. returns indices for these scores

        # step 3: compute background
        background_matrix = matrix[i_background["background"]]
        bg_labels = labels[i_background["background"]]
        if len(background_matrix) > 0:
            bg_probas = feature_class_probabilities(
                background_matrix.T, bg_labels, kmers=self.kmers.basis
            )
            self.probabilities["background"] = bg_probas[
                bg_probas["label"] == self.primary_label
            ]["score"].values

            # step 3.1: background family probability scores
            for bl in np.unique(bg_labels):
                bg_scores = bg_probas[bg_probas["label"] == bl]["score"].values

                # normalize by max bg score
                # bg_norm = np.max(bg_scores)

                # get background scores for sequences
                bg_only_scores = apply_feature_probabilities(
                    matrix, bg_scores, scaler=self.scaler
                )

                # normalize by max bg score attained by a sequence
                bg_norm = np.max(bg_only_scores)

                self.scores["background"][bl] = bg_only_scores / bg_norm

        # step 3.2: assign family probability scores
        for sl in np.unique(s_labels):
            scores = probas[probas["label"] == sl]["score"].values

            # normalize by sum of all positive scores
            # norm = np.sum([s for s in scores if s > 0])

            # include background sequences for score generation
            total_scores = apply_feature_probabilities(
                matrix, scores, scaler=self.scaler
            )

            # normalize by max score
            norm = np.max(total_scores)
            norm = max(norm, 1.0)  # prevent score explosion?
            if sl == self.primary_label:
                self.score_norm = norm

            # assign percent score based on max positive score
            self.scores["sample"][f"{sl}_score"] = total_scores / norm

            # weight family score by (1 - normalized bg score)
            if sl in np.unique(bg_labels):
                self.scores["sample"][f"{sl}_score_background_weighted"] = [
                    total * (1 - bg)
                    for total, bg in zip(
                        self.scores["sample"][f"{sl}_score"],
                        self.scores["background"][sl],
                    )
                ]

                # old scoring method
                self.scores["sample"][f"{sl}_background_subtracted_score"] = [
                    total - bg
                    for total, bg in zip(
                        self.scores["sample"][f"{sl}_score"],
                        self.scores["background"][sl],
                    )
                ]

    # score new input sequences
    def predict(self, array, kmers):
        """Predict scores of new array with separate kmer basis.

        Parameters
        ----------
        array : list or array-like
            Description of parameter `array`.
        kmers : list or array-like of str
            Ordered array of kmer strings.

        Returns
        -------
        type
            Description of returned object.

        """
        # fit new input data to the correct kmer basis
        # self.kmers.set_basis(kmers)
        array = self.kmers.transform(array, kmers)

        # return (
        #    apply_feature_probabilities(
        #        array, self.probabilities["sample"], scaler=self.scaler
        #    )
        #    / self.score_norm
        # )
        return (
            apply_feature_probabilities(
                array, self.probabilities["sample"], scaler=False
            )
            / self.score_norm
        )



"""io: Input-output handling in Snekmer.

author: @christinehc

"""
# imports
import gzip
import json
import pickle
import re
from ast import literal_eval
from os.path import basename, join, splitext
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from .alphabet import ALPHABET_ORDER


# functions
def load_pickle(filename: str, mode: str = "rb") -> Any:
    """Load a pickled object (wrapper for `pickle.load`).

    Parameters
    ----------
    filename : str
        Description of parameter `filename`.
    mode : str
        Description of parameter `mode` (the default is "rb").

    Returns
    -------
    Any
        Description of returned object.

    Raises
    ------
    ExceptionName
        Why the exception is raised.

    """
    with open(filename, mode) as f:
        return pickle.load(f)


def load_npz(
    filename: str,
    columns: Dict[str, str] = {
        "ids": "sequence_id",
        "seqs": "sequence",
        "vecs": "sequence_vector",
    },
    objects: Tuple = ("kmerlist",),
) -> Tuple[List, pd.DataFrame]:
    """Compile .npz results into dataframe.

    Parameters
    ----------
    filename : str
        /path/to/filename for input .npz file.
    columns : Dict[str, str]
        Mapping for output data column names (the default is
            {
                "ids": "sequence_id",
                "seqs": "sequence",
                "vecs": "sequence_vector",
            }
        ).
    objects : Tuple[str]
        Column names for additional objects to return

    Returns
    -------
    Tuple[List, pd.DataFrame]
        Tuple with list of data objects and tabulated .npz data.

    """
    data = np.load(filename)

    # fill in df based on desired output col names
    df = {"filename": splitext(basename(filename))[0]}

    for in_col, out_col in columns.items():
        df.update({out_col: list(data[in_col])})

        # get seq column for sequence lengths
        if "seq" in in_col:
            df.update({f"{out_col}_length": [len(s) for s in data[in_col]]})

    # load things that don't fit in the data matrix format
    #    such as kmerbasis
    extras = list()
    for obj in objects:
        extras.append(data[obj])

    return extras, pd.DataFrame(df)


def read_kmers(filename: str) -> List[str]:
    """Extract kmer identifiers from Snekmer output file.

    Parameters
    ----------
    filename : str
        /path/to/output/file.txt

    Returns
    -------
    list
        Array of strings (kmer identifiers)

    """
    kmers = list()
    with open(filename) as f:
        for line in f:
            kmers.append(line.strip())  # .split("\t")
            # parse kmer outputs if detected
            # if re.findall(r"^KMER", line_data[0]):
            # kmers += line_data
    # prefix = re.search(r"^KMER-[\d]+-[A-Za-z]+-", kmers[0]).group()
    # return [s.strip().replace(prefix, "") for s in kmers]
    return kmers


def vecfiles_to_df(
    files: List[str], labels: Optional[List[str]] = None, label_name: str = "label"
) -> pd.DataFrame:
    """Load multiple sequence files and parse into common dataframe.

    Parameters
    ----------
    files : list of str
        List of (/path/to/file1.ext1, /path/to/file2.ext2, ..., etc.)
    labels : list of str or None (default: None)
        Labels for specified files; if None, excludes labels.
    label_name : str (default: 'label')
        Desired name for data column containing labels.

    Returns
    -------
    pandas.DataFrame
        Tabulated data of the form:

        | filename | seq_id | vector | vec_shape |
        |----------|--------|--------|-----------|
        | file1    | seq1   | array1 | shape1    |
        ...
        ... etc.

    """
    data = {"filename": [], "seq_id": [], "vector": [], "vec_shape": []}
    for afile in files:
        with gzip.open(afile, "rt") as f:
            # with open(afile, 'r') as f:
            tmp = json.load(f)
            data["filename"] += [basename(afile)] * len(tmp["seq_id"])
            data["seq_id"] += tmp["seq_id"]
            data["vector"] += tmp["vector"]
            data["vec_shape"] += [np.array(arr).shape for arr in tmp["vector"]]
    if str(labels) != "None":
        if len(labels) != len(data):
            raise ValueError("Number of labels must equal number of sequences.")
        data[label_name] = labels
    return pd.DataFrame(data)


def define_output_dir(alphabet: Union[str, int], k: int, nested: bool = False) -> str:
    """Create output directory name using AAR parameters.

    Parameters
    ----------
    alphabet : str or int
        Name or numerical identifier of alphabet.
        See `snekmer.alphabet.ALPHABETS` for descriptions.
    k : int
        Kmer length.
    nested : bool (default: False)
        If True, creates nested directories with alphabet, k.
        If False, returns simple output directory.

    Returns
    -------
    str
        Name of output directory, given nested directory parameters.

    """
    if isinstance(nested, str):
        nested = literal_eval(nested)
    if not nested:
        return "output"
    if not isinstance(alphabet, str):
        alphabet = ALPHABET_ORDER[alphabet]
    return join("output", alphabet, f"k-{k:02}")



# ---------------------------------------------------------
# Imports
# ---------------------------------------------------------

from datetime import datetime
from os import makedirs
from os.path import dirname, exists, join

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import snekmer as skm
from scipy.spatial.distance import squareform
from scipy.spatial.distance import pdist
from umap import UMAP

# ---------------------------------------------------------
# (optional) BSF handling
# ---------------------------------------------------------

try:
    # make this conditional on the library being installed
    import bsf

    BSF_PRESENT = True

except ImportError:
    BSF_PRESENT = False

# ---------------------------------------------------------
# Files and Parameters
# ---------------------------------------------------------

config = snakemake.config

# change matplotlib backend to non-interactive
plt.switch_backend("Agg")

# ---------------------------------------------------------
# Run script
# ---------------------------------------------------------

# log script start time
start_time = datetime.now()
with open(snakemake.log[0], "a") as f:
    f.write(f"start time:\t{start_time}\n")

# parse all data and label background files
label = config["score"]["lname"]

# assuming all kmer basis sets are identical, grab the first
kmer = skm.io.load_pickle(snakemake.input.kmerobj[0])

# tabulate vectorized seq data
data, kmers = list(), list()
for dfile, kfile in zip(sorted(snakemake.input.data), sorted(snakemake.input.kmerobj)):
    kmerlist, df = skm.io.load_npz(dfile)
    data.append(df)

    kobj = skm.io.load_pickle(kfile)
    klist = kobj.kmer_set._kmerlist
    kmers.append(klist)
# basis.extend(klist)

# make a superset of kmers
kmerbasis = np.unique(np.hstack(kmers))

basis = skm.vectorize.KmerVec(config["alphabet"], config["k"])
basis.set_kmer_set(kmerbasis)

for i in range(len(data)):
    df, kmerlist = data[i], kmers[i]
    vecs = skm.utils.to_feature_matrix(df["sequence_vector"].values)

    converted = basis.harmonize(vecs, kmerlist)
    df["sequence_vector"] = converted.tolist()

data = pd.concat(data, ignore_index=True)
data["background"] = [f in snakemake.input.bg for f in data["filename"]]

# log conversion step runtime
skm.utils.log_runtime(snakemake.log[0], start_time, step="load_npz")

# define feature matrix of kmer vectors not from background set
bg, non_bg = data[data["background"]], data[~data["background"]]
full_feature_matrix = skm.utils.to_feature_matrix(data["sequence_vector"].values)
feature_matrix = skm.utils.to_feature_matrix(non_bg["sequence_vector"].values)

# filter by kmer occurrence
# print("Filtering?")
if str(config["cluster"]["min_rep"]) != "None":
    min_rep = max_rep = config["cluster"]["min_rep"]
    print(f"Filtering by minimum {min_rep}")
    ksums = full_feature_matrix.sum(axis=0)
    full_feature_matrix = full_feature_matrix[:, ksums > min_rep]
    print(full_feature_matrix.shape)

if str(config["cluster"]["max_rep"]) != "None":
    max_rep = config["cluster"]["max_rep"]
    print(f"Filtering by maximum {max_rep}")
    ksums = full_feature_matrix.sum(axis=0)
    full_feature_matrix = full_feature_matrix[:, ksums < max_rep]
    print(full_feature_matrix.shape)

# currently not used
# if len(bg) > 0:
#     bg_feature_matrix = skm.utils.to_feature_matrix(bg["sequence_vector"].values)
# kludge to check on options for agglomerative Clustering
if (
    "n_clusters" in config["cluster"]["params"]
    and config["cluster"]["params"]["n_clusters"] == "None"
):
    config["cluster"]["params"]["n_clusters"] = None
    config["cluster"]["params"]["compute_full_tree"] = True

# initialize clustering model
model = skm.cluster.KmerClustering(
    config["cluster"]["method"], config["cluster"]["params"]
)

# for bsf, make the input matrix a distance matrix
if config["cluster"]["method"] in [
    "density-jaccard",
    "hdensity-jaccard",
    "agglomerative-jaccard",
]:
    # print BSF status
    output_bsf = join(dirname(snakemake.output.table), "bsf")
    if not exists(output_bsf):
        makedirs(output_bsf)

    # if available, use BSF to create clustering distance matrix
    if BSF_PRESENT:
        print("Using BSF to compute distance matrix...")
        full_feature_matrix = np.array(
            full_feature_matrix, dtype=np.uint64
        )  # bsf requires uint64

        # the filename that bsf creates you can't specify exactly
        # but we can reconstruct it here - this will likely break
        # because I think bsf breaks large input matrices into chunks
        # and will output multiple files
        nsign, vlen = full_feature_matrix.shape

        bsfname = join(
            output_bsf, "bin_%d_0_0_%d_%d_bsf.txt.bin" % (nsign, nsign, nsign)
        )

        if not exists(bsfname):
            # We can break this job in to chunks by setting the nsign to smaller
            #    than the row size - but we'll need to figure out how to read
            #    in the bin files and use them
            # bsf.analysis_with_chunk(full_feature_matrix, 10000, "bsf.txt", "%s/" % output.bsf)
            bsf.analysis_with_chunk(
                full_feature_matrix, nsign, "bsf.txt", "%s/" % output_bsf
            )
        else:
            print("Using existing BSF similarity matrix...")

        with open(bsfname, "rb") as f:
            bsf_mat = np.frombuffer(f.read(), dtype=np.uint32).reshape((nsign, nsign))
        # bsf_mat = np.frombuffer(f.read(), dtype=np.double).reshape((nsign, nsign))

        # this is great for diagnostics - but takes a lot of time/space for large
        # comparisons
        # np.savetxt(output.bsfmat, bsf_mat, fmt="%.3f", delimiter=",")

        # bsf only outputs the upper triangle so we'll copy that to the lower triangle
        bsf_mat = bsf_mat + bsf_mat.T - np.diag(np.diag(bsf_mat))

        # and fill the diagonals with max which is skipped by bsf
        np.fill_diagonal(bsf_mat, 100)

        # now transform the similarity matrix output by bsf to a
        #     distance matrix - using inverse Jaccard similarity

        # this gives Jaccard similarity (since all vectors are the same length)
        # and subtracting it from 1 gives a distance
        # bsf_mat = 1.0 - (bsf_mat / 100)
        bsf_mat = (100 - bsf_mat) / 100.0

    else:
        # If bsf isn't available we'll use scipy to
        # calculate jaccard distance in a similar way
        # but it will be slower and more memory intensive
        # especially for larger clustering jobs
        # calculate Jaccard distance
        res = pdist(full_feature_matrix, "jaccard")
        bsf_mat = squareform(res)

    # dist_thresh = config["cluster"]["dist_thresh"]
    # bsf_mat[bsf_mat > dist_thresh] = 100

    # output matrix for diagnostics - time/space heavy for large comparisons
    if config["cluster"]["save_matrix"] is True:
        np.savetxt(
            join(output_bsf, "bsf_matrix.csv"), bsf_mat, fmt="%.3f", delimiter=",",
        )

    model.fit(bsf_mat)
else:
    # fit and save fitted clusters
    # bsf here
    model.fit(full_feature_matrix)

# save output
data["cluster"] = model.labels_
data = data.drop(columns=["sequence", "sequence_vector"])
data.to_csv(snakemake.output.table, index=False)

# with open(output.clusters, "wb") as f:
#     pickle.dump(model, f)

# always create output figure directory
if not exists(snakemake.output.figs):
    makedirs(snakemake.output.figs)

# log time to compute clusters
skm.utils.log_runtime(snakemake.log[0], start_time, step="clustering")

# optionally generate plots
if str(config["cluster"]["cluster_plots"]) == "True":
    # plot explained variance curve
    fig, ax = skm.plot.explained_variance_curve(full_feature_matrix)
    fig.savefig(join(snakemake.output.figs, "pca_explained_variance_curve.png"))
    plt.close("all")

    # plot tsne
    fig, ax = skm.plot.cluster_tsne(full_feature_matrix, model.model.labels_)
    fig.savefig(join(snakemake.output.figs, "tsne.png"))
    plt.close("all")

    # plot umap
    umap_embedding = UMAP(metric="jaccard", n_components=2).fit_transform(
        full_feature_matrix
    )
    fig, ax = plt.subplots(dpi=150)
    sns.scatterplot(
        x=umap_embedding[:, 0],
        y=umap_embedding[:, 1],
        hue=model.model.labels_,
        alpha=0.2,
        ax=ax,
    )

    fig.savefig(join(snakemake.output.figs, "umap.png"))
    plt.close("all")

# record script endtime
skm.utils.log_runtime(snakemake.log[0], start_time)



"""cluster: Snekmer clustering models.

author: @christinehc

"""
# imports
import numpy as np
from hdbscan import HDBSCAN
from sklearn.base import ClusterMixin
from sklearn.cluster import (
    AgglomerativeClustering,
    DBSCAN,
    Birch,
    OPTICS,
    MiniBatchKMeans,
)
from scipy.cluster.hierarchy import fclusterdata

# from umap import UMAP


# wrap scipy correlation clustering into sklearn-like API
class CorrelationClustering(ClusterMixin):
    """Cluster input by correlation using hierarchical clustering.

    Based on `scipy.hierarchy.fcluster`.

    Parameters
    ----------
    **model_params : dict or params
        Parameters for model. See documentation for
        `scipy.hierarchy.fcluster`.

    Attributes
    ----------
    model_params : dict
        Model initialization parameters.
    labels_ : list
        Cluster-assigned labels.

    """

    def __init__(self, **model_params):
        self.model_params = dict(model_params)
        self.labels_ = None

    def fit(self, X):
        self.labels_ = fclusterdata(X, **self.model_params)

    def predict(self, X):
        return fclusterdata(X, **self.model_params)


# wrap scipy correlation clustering into sklearn-like API
class BSFClustering(ClusterMixin):
    """Cluster input using BSF as distance matrix

    Parameters
    ----------
    **model_params : dict or params
        Parameters for model.

    Attributes
    ----------
    model_params : dict
        Model initialization parameters.
    labels_ : list
        Cluster-assigned labels.

    """

    def __init__(self, **model_params):
        model_params["metric"] = "precomputed"
        self.model_params = dict(model_params)
        self.labels_ = None
        self.method = DBSCAN(**model_params)

    def fit(self, X):
        self.method.fit(X)
        self.labels_ = np.full(X.shape[0], -1, dtype=np.intp)

    def predict(self, X):
        return self.method.fit_predict(X)


class HDBSFClustering(BSFClustering):
    def __init__(self, **model_params):
        model_params["metric"] = "precomputed"
        self.model_params = dict(model_params)
        self.labels_ = None
        self.method = HDBSCAN(**model_params)


class BSFAgglomerative(BSFClustering):
    def __init__(self, **model_params):
        model_params["affinity"] = "precomputed"
        self.model_params = dict(model_params)
        self.labels_ = None
        self.method = AgglomerativeClustering(**model_params)

    def fit(self, X):
        self.method.fit(X)
        self.labels_ = self.method.labels_

# define allowed model types
MODELS = {
    "kmeans": MiniBatchKMeans,
    "agglomerative": AgglomerativeClustering,
    "correlation": CorrelationClustering,
    "density": DBSCAN,
    "birch": Birch,
    "optics": OPTICS,
    "density-jaccard": BSFClustering,
    "hdensity-jaccard": HDBSFClustering,
    "agglomerative-jaccard": BSFAgglomerative,
    "hdbscan": HDBSCAN,
}


# wrapper class for all clustering model types
class KmerClustering(ClusterMixin):
    """Cluster kmer data using various metrics.

    Parameters
    ----------
    method : str
        Clustering method; see `snekmer.cluster.MODELS` for details.
    model_params : type
        Description of parameter `model_params`.

    Attributes
    ----------
    model : type
        Description of attribute `model`.
    method

    """

    def __init__(self, method, model_params={}):
        self.method = method
        self.model = MODELS[method](**model_params)

    def fit(self, X):
        self.model.fit(X)
        if self.method != "hdbscan":
            #self.labels_ = np.full(X.shape[0], -1, dtype=np.intp)
            self.labels_ = self.model.labels_
        else:
            self.labels_ = self.model.fit_predict(X)

    def predict(self, X):
        if self.method != "hdbscan":
            return self.model.predict(X)
        raise AttributeError("'HDBSCAN' object has no attribute 'predict'.")



"""plot: Snekmer plotting utilities.

author: @christinehc

"""
# imports
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from numpy.typing import ArrayLike
from sklearn.metrics import (
    PrecisionRecallDisplay,
    RocCurveDisplay,
    auc,
    average_precision_score,
    precision_recall_curve,
)
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE


def cv_roc_curve(
    clf,
    X: ArrayLike,
    y: ArrayLike,
    title: str = "ROC Curve",
    ax: Optional[plt.Axes] = None,
    dpi: int = 400,
):
    """Plot cross-validated receiver operator characteristics curve.

    Adapted from example in sklearn documentation [1].
    [1]: http://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html#sphx-glr-auto-examples-model-selection-plot-roc-crossval-py

    Parameters
    ----------
    clf : object
        Classifier object (e.g. DecisionTreeClassifier()).
    X : numpy.typing.ArrayLike
        Feature array.
    y : numpy.typing.ArrayLike
        Response array.
    title : str
        Plot title.
    ax : matplotlib.axes.Axes or None (default: None)
        Axis object, if already created.
    dpi : int (default: 400)
        Figure resolution.

    Returns
    -------
    fig, ax, auc_rocs

    """
    # initialize figure and define all axes
    fig = plt.figure(
        dpi=dpi, figsize=(8, 4), facecolor="white", constrained_layout=True
    )
    ax = ax or plt.gca()

    # create empty arrays to store data
    tprs, auc_rocs = [], []
    mean_fpr = np.linspace(0, 1, 100)

    # take each cv result
    for i in X.keys():
        clf.fit(X[i]["train"], y[i]["train"])
        viz = RocCurveDisplay.from_estimator(
            clf.model,
            X[i]["test"],
            y[i]["test"],
            name=f"ROC fold {i}",
            alpha=0.3,
            lw=1,
            ax=ax,
        )
        interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
        interp_tpr[0] = 0.0
        tprs.append(interp_tpr)
        auc_rocs.append(viz.roc_auc)

    # plot random chance reference line
    ax.plot([0, 1], [0, 1], "k--", label="Chance level (AUC = 0.5)")

    # calculate mean and stdev roc_auc and show in plot
    mean_tpr = np.mean(tprs, axis=0)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    std_auc = np.std(auc_rocs)
    ax.plot(
        mean_fpr,
        mean_tpr,
        color="b",
        label=f"Mean ROC (AUC = {mean_auc:0.2f} ± {std_auc:0.2f})",
        lw=2,
        alpha=0.8,
    )

    # draw shaded area for +- 1 stdev from mean auc roc
    std_tpr = np.std(tprs, axis=0)
    tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
    tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
    ax.fill_between(
        mean_fpr,
        tprs_lower,
        tprs_upper,
        color="grey",
        alpha=0.2,
        label=r"$\pm$ 1 std. dev.",
    )

    # set fixed plot limits from ~0 to ~1
    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])
    ax.set_xlabel("False Positive Rate")
    ax.set_ylabel("True Positive Rate")
    ax.set_title(title)
    ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0.0)

    return fig, ax, auc_rocs


def cv_pr_curve(
    clf,
    X: ArrayLike,
    y: ArrayLike,
    title: str = "PR Curve",
    ax: Optional[plt.Axes] = None,
    dpi: int = 400,
):
    """Plot cross-validated precision-recall curve.

    Parameters
    ----------
    clf : Classifer object
        Classifier object (e.g. DecisionTreeClassifier()).
    X : numpy.typing.ArrayLike
        Feature array.
    y : numpy.typing.ArrayLike
        Response array.
    title : str
        Plot title.
    ax : matplotlib.axes.Axes or None (default: None)
        Axis object, if already created.
    dpi : int (default: 400)
        Figure resolution.

    Returns
    -------
    fig, ax, pr_aucs

    """
    # initialize figure and define all axes
    fig = plt.figure(
        dpi=dpi, figsize=(8, 4), facecolor="white", constrained_layout=True
    )
    ax = ax or plt.gca()

    # create empty arrays to store data
    y_real, y_proba, pr_aucs = [], [], []

    for i in X.keys():
        clf.fit(X[i]["train"], y[i]["train"])
        probabilities = clf.model.predict_proba(X[i]["test"])
        precision, recall, _ = precision_recall_curve(
            y[i]["test"], clf.predict(X[i]["test"])
        )
        viz = PrecisionRecallDisplay.from_estimator(
            clf.model,
            X[i]["test"],
            y[i]["test"],
            name=f"ROC fold {i}",
            alpha=0.3,
            lw=1,
            ax=ax,
        )

        y_real.append(y[i]["test"])
        y_proba.append(probabilities[:, 1])
        pr_aucs.append(viz.average_precision)

    y_real = np.concatenate(y_real)
    y_proba = np.concatenate(y_proba)

    precision, recall, _ = precision_recall_curve(y_real, y_proba)
    # avg_total_precision = average_precision_score(y_real, y_proba)

    # plot average p-r curve
    ax.plot(
        recall,
        precision,
        color="b",
        label=f"Precision-Recall (AUC = {viz.average_precision:0.2f})",
        lw=2,
        alpha=0.8,
    )

    # set fixed plot limits from ~0 to ~1
    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])
    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    ax.set_title(title)
    ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0.0)
    return fig, ax, pr_aucs


# plot PCA explained variance
def explained_variance_curve(feature_matrix):
    X_scaled = StandardScaler().fit_transform(feature_matrix)
    pca = PCA()
    pca.fit(X_scaled)

    fig, ax = plt.subplots(dpi=200, figsize=(4, 3), constrained_layout=True)
    ax.plot(
        [0] + list(np.arange(1, pca.n_components_ + 1)),
        [0] + list(np.cumsum(pca.explained_variance_ratio_)),
    )
    ax.set_ylabel("PCA explained variance ratio")

    return fig, ax


# plot clusters using tsne (requires scaling + pca)
def cluster_tsne(feature_matrix, clusters, **tsne_args):
    X_scaled = StandardScaler().fit_transform(feature_matrix)
    X_pca = PCA().fit_transform(X_scaled)
    X_embedded = TSNE(**tsne_args).fit_transform(X_pca)  # [:, :100])  #_unscaled)

    fig, ax = plt.subplots(dpi=200, figsize=(4, 3), constrained_layout=True)
    sns.scatterplot(x=X_embedded[:, 0], y=X_embedded[:, 1], hue=clusters, ax=ax)
    return fig, ax



"""cli: Command line interface for Snekmer.

author: @christinehc, @biodataganache, @snakemake

"""
# imports
import argparse
import os

from multiprocessing import cpu_count
from pkg_resources import resource_filename
from snakemake import snakemake, parse_config, get_profile_file
from snekmer import __version__

# define options
MAP_FN_DESC = [
    "Hydrophobic/hydrophilic",
    "Standard 7",
    "Solvent accessibility",
    "LoveHateCharge",
    "LoveHateBadstruct",
]
MAP_FNS = {f"reduced_alphabet_{n}": MAP_FN_DESC[n] for n in range(5)}


def get_argument_parser():
    parser = {}

    # main parser
    parser["main"] = argparse.ArgumentParser(
        description=(
            "Snekmer: A tool for kmer-based sequence analysis"
            " using amino acid reduction (AAR)."
        )
    )
    parser["main"].add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        help="Print version and exit.",
    )

    # Snakemake options
    # NOTE: Arguments and documentation borrowed from Snakemake:
    # https://github.com/snakemake/snakemake/blob/main/snakemake/__init__.py
    parser["smk"] = argparse.ArgumentParser(add_help=False)
    parser["smk"].add_argument(
        "--dry-run",
        "--dryrun",
        "-n",
        dest="dryrun",
        action="store_true",
        help="Do not execute anything, and display what would be done. "
        "If you have a very large workflow, use --dry-run --quiet to just "
        "print a summary of the DAG of jobs.",
    )
    parser["smk"].add_argument(
        "--configfile",
        nargs="+",
        # default="config.yaml",
        metavar="PATH",
        help=(
            "Specify or overwrite the config file of the workflow (see the docs). "
            "Values specified in JSON or YAML format are available in the global config "
            "dictionary inside the workflow. Multiple files overwrite each other in "
            "the given order. Thereby missing keys in previous config files are extended by "
            "following configfiles. Note that this order also includes a config file defined "
            "in the workflow definition itself (which will come first)."
        ),
    )
    parser["smk"].add_argument(
        "-C",
        "--config",
        nargs="*",
        metavar="KEY=VALUE",
        help=(
            "Set or overwrite values in the"
            " workflow config object. The workflow"
            " config object is accessible as"
            " variable config inside the workflow."
            " Default values can be set by"
            " providing a JSON file."
        ),
    )
    parser["smk"].add_argument("--unlock", action="store_true", help="unlock directory")
    parser["smk"].add_argument(
        "--until",
        "-U",
        nargs="+",
        metavar="TARGET",
        help=(
            "Runs the pipeline until it reaches the specified rules or "
            "files. Only runs jobs that are dependencies of the specified "
            "rule or files, does not run sibling DAGs. "
        ),
    )
    parser["smk"].add_argument(
        "--keepgoing",
        "--keep-going",
        "-k",
        action="store_true",
        help="Go on with independent jobs if a job fails.",
    )
    parser["smk"].add_argument(
        "--latency",
        "-w",
        "--output-wait",
        "--latency-wait",
        type=int,
        default=30,
        metavar="SECONDS",
        help="Wait given seconds if an output file of a job is not present after "
        "the job finished. This helps if your filesystem "
        "suffers from latency (default 30).",
    )
    parser["smk"].add_argument(
        "--touch",
        "-t",
        action="store_true",
        help=(
            "Touch output files (mark them up to date without really "
            "changing them) instead of running their commands. This is "
            "used to pretend that the rules were executed, in order to "
            "fool future invocations of snakemake. Fails if a file does "
            "not yet exist. Note that this will only touch files that would "
            "otherwise be recreated by Snakemake (e.g. because their input "
            "files are newer). For enforcing a touch, combine this with "
            "--force, --forceall, or --forcerun. Note however that you loose "
            "the provenance information when the files have been created in "
            "realitiy. Hence, this should be used only as a last resort."
        ),
    )
    parser["smk"].add_argument(
        "--cores",
        "-c",
        default=cpu_count(),
        type=int,
        metavar="N",
        help=(
            "Use at most N CPU cores/jobs in parallel. "
            "If N is omitted or 'all', the limit is set to the number of "
            "available CPU cores. "
            "In case of cluster/cloud execution, this argument sets the maximum number "
            "of cores requested from the cluster or cloud scheduler. (See "
            "https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#"
            "resources-remote-execution for more info)"
            "This number is available to rules via workflow.cores."
        ),
    )
    parser["smk"].add_argument(
        "--count",
        metavar="N",
        type=int,
        help="Number of files to process (limits DAG size).",
    )
    parser["smk"].add_argument(
        "--countstart",
        metavar="IDX",
        type=int,
        default=0,
        help="Starting file index (for use with --count).",
    )
    parser["smk"].add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="Show additional debug output (default False)",
    )
    parser["smk"].add_argument(
        "--quiet",
        "-q",
        nargs="*",
        choices=["progress", "rules", "all"],
        default=None,
        help=(
            "Do not output certain information. If used without "
            "arguments, do not output any progress or rule "
            "information. Defining 'all' results in no "
            "information being printed at all."
        ),
    )
    parser["smk"].add_argument(
        "--directory",
        "-d",
        metavar="DIR",
        action="store",
        help=(
            "Specify working directory (relative paths in "
            "the snakefile will use this as their origin)."
        ),
    )
    parser["smk"].add_argument(
        "--forcerun",
        "-R",
        nargs="*",
        metavar="TARGET",
        help=(
            "Force the re-execution or creation of the given rules or files."
            " Use this option if you changed a rule and want to have all its "
            "output in your workflow updated."
        ),
    )
    parser["smk"].add_argument(
        "--list-code-changes",
        "--lc",
        action="store_true",
        help="List all output files for which the rule body (run or shell) have "
        "changed in the Snakefile.",
    )
    parser["smk"].add_argument(
        "--list-params-changes",
        "--lp",
        action="store_true",
        help="List all output files for which the defined params have changed "
        "in the Snakefile.",
    )

    # clust execution options
    parser["clust"] = parser["smk"].add_argument_group("Cluster Execution Arguments")
    parser["clust"].add_argument(
        "--clust",
        nargs="+",
        metavar="PATH",
        help="Path to cluster execution yaml configuration file.",
    )
    parser["clust"].add_argument(
        "-j",
        "--jobs",
        metavar="N",
        type=int,
        default=1000,
        help="Number of simultaneous jobs to submit to a slurm queue.",
    )

    # create subparsers for each operation mode
    # parser.add_argument("mode", choices=["cluster", "model", "search", "learn", "apply"])
    parser["subparsers"] = parser["main"].add_subparsers(
        title="mode",
        description="Snekmer mode (cluster, model, search, learn, or apply).",
        dest="mode",
    )

    # subparsers
    parser["cluster"] = parser["subparsers"].add_parser(
        "cluster",
        description="Apply unsupervised clustering via Snekmer.",
        parents=[parser["smk"]],
    )
    parser["model"] = parser["subparsers"].add_parser(
        "model",
        description="Train supervised models via Snekmer.",
        parents=[parser["smk"]],
    )
    parser["search"] = parser["subparsers"].add_parser(
        "search",
        description="Search sequences against pre-existing models via Snekmer.",
        parents=[parser["smk"]],
    )
    parser["learn"] = parser["subparsers"].add_parser(
        "learn",
        description="Learn kmer-annotation associations via Snekmer",
        parents=[parser["smk"]],
    )
    parser["apply"] = parser["subparsers"].add_parser(
        "apply",
        description="Apply kmer-annotation associations via Snekmer",
        parents=[parser["smk"]],
    )
    return parser


def get_main_args():
    parser = get_argument_parser()
    return parser["main"]


def main():
    parser = get_argument_parser()

    # parse args
    args = parser["main"].parse_args()
    config = parse_config(args)

    # start/stop config
    if args.count is not None:
        config = {
            "start": args.countstart,
            "stop": args.countstart + args.count,
            **config,
        }
    else:
        config = config

    # set quiet output options
    if args.quiet is not None and len(args.quiet) == 0:
        # default case, set quiet to progress and rule
        args.quiet = ["progress", "rules"]

    # cluster config
    if args.clust is not None:
        cluster = "sbatch -A {cluster.account} -N {cluster.nodes} -t {cluster.time} -J {cluster.name} --ntasks-per-node {cluster.ntasks} -p {cluster.partition}"
    else:
        cluster = None

    # fix configfile path
    if args.configfile is None:
        configfile = ["config.yaml"]
    else:
        configfile = args.configfile

    if (args.directory is not None) and (args.configfile is None):
        configfile = [os.path.join(args.directory, c) for c in configfile]
    else:
        configfile = list(map(os.path.abspath, configfile))

    # parse operation mode
    if args.mode == "cluster":
        snakemake(
            resource_filename("snekmer", os.path.join("rules", "cluster.smk")),
            configfiles=configfile,
            config=config,
            cluster_config=args.clust,
            cluster=cluster,
            keepgoing=args.keepgoing,
            force_incomplete=True,
            forcerun=args.forcerun,
            cores=args.cores,
            nodes=args.jobs,
            workdir=args.directory,
            dryrun=args.dryrun,
            unlock=args.unlock,
            list_code_changes=args.list_code_changes,
            list_params_changes=args.list_params_changes,
            until=args.until,
            touch=args.touch,
            latency_wait=args.latency,
            verbose=args.verbose,
            quiet=args.quiet,
        )

    elif args.mode == "model":
        snakemake(
            resource_filename("snekmer", os.path.join("rules", "model.smk")),
            configfiles=configfile,
            config=config,
            cluster_config=args.clust,
            cluster=cluster,
            keepgoing=args.keepgoing,
            force_incomplete=True,
            forcerun=args.forcerun,
            cores=args.cores,
            nodes=args.jobs,
            workdir=args.directory,
            dryrun=args.dryrun,
            unlock=args.unlock,
            list_code_changes=args.list_code_changes,
            list_params_changes=args.list_params_changes,
            until=args.until,
            touch=args.touch,
            latency_wait=args.latency,
            verbose=args.verbose,
            quiet=args.quiet,
        )

    elif args.mode == "search":
        snakemake(
            resource_filename("snekmer", os.path.join("rules", "search.smk")),
            configfiles=configfile,
            config=config,
            cluster_config=args.clust,
            cluster=cluster,
            keepgoing=True,
            force_incomplete=True,
            forcerun=args.forcerun,
            cores=args.cores,
            nodes=args.jobs,
            workdir=args.directory,
            dryrun=args.dryrun,
            unlock=args.unlock,
            list_code_changes=args.list_code_changes,
            list_params_changes=args.list_params_changes,
            until=args.until,
            touch=args.touch,
            latency_wait=args.latency,
            verbose=args.verbose,
            quiet=args.quiet,
        )

    elif args.mode == "learn":
        snakemake(
            resource_filename("snekmer", os.path.join("rules", "learn.smk")),
            configfiles=configfile,
            config=config,
            cluster_config=args.clust,
            cluster=cluster,
            keepgoing=True,
            force_incomplete=True,
            forcerun=args.forcerun,
            cores=args.cores,
            nodes=args.jobs,
            workdir=args.directory,
            dryrun=args.dryrun,
            unlock=args.unlock,
            list_code_changes=args.list_code_changes,
            list_params_changes=args.list_params_changes,
            until=args.until,
            touch=args.touch,
            latency_wait=args.latency,
            verbose=args.verbose,
            quiet=args.quiet,
        )
        
    elif args.mode == "apply":
        snakemake(
            resource_filename("snekmer", os.path.join("rules", "apply.smk")),
            configfiles=configfile,
            config=config,
            cluster_config=args.clust,
            cluster=cluster,
            keepgoing=True,
            force_incomplete=True,
            forcerun=args.forcerun,
            cores=args.cores,
            nodes=args.jobs,
            workdir=args.directory,
            dryrun=args.dryrun,
            unlock=args.unlock,
            list_code_changes=args.list_code_changes,
            list_params_changes=args.list_params_changes,
            until=args.until,
            touch=args.touch,
            latency_wait=args.latency,
            verbose=args.verbose,
            quiet=args.quiet,
        )

    else:
        parser["main"].print_help()


if __name__ == "__main__":
    main()




# ---------------------------------------------------------
# Imports
# ---------------------------------------------------------

import pickle
from datetime import datetime
from os import makedirs
from os.path import exists, join

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import snekmer as skm
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split, StratifiedKFold
from sklearn.preprocessing import LabelEncoder

# ---------------------------------------------------------
# Files and Parameters
# ---------------------------------------------------------

config = snakemake.config

# change matplotlib backend to non-interactive
plt.switch_backend("Agg")

# ---------------------------------------------------------
# Run script
# ---------------------------------------------------------

# create lookup table to match vector to sequence by file+ID
# lookup = {}
# for f in input.raw:
#    loaded = np.load(f)
#    lookup.update(
#        {
#            (splitext(basename(f))[0], seq_id): seq_vec
#            for seq_id, seq_vec in zip(loaded["ids"], loaded["vecs"])
#        }
#    )
# print(lookup)
with open(snakemake.input.matrix, "rb") as f:
    data = pickle.load(f)

# load all input data and encode rule-wide variables
# data = pd.read_csv(input.data)
# data["sequence_vector"] = [
#    lookup[(seq_f, seq_id)]
#    for seq_f, seq_id in zip(data["filename"], data["sequence_id"])
# ]

scores = pd.read_csv(snakemake.input.weights)
family = skm.utils.get_family(
    skm.utils.split_file_ext(snakemake.input.weights)[0],
    regex=config["input_file_regex"],
)
# get kmers for this particular set of sequences
with open(snakemake.input.kmerobj, "rb") as f:
    kmer = pickle.load(f)

cv = config["model"]["cv"]

# set category label name (e.g. "family")
label = config["score"]["lname"] if str(config["score"]["lname"]) != "None" else "label"

# prevent kmer NA being read as np.nan
if config["k"] == 2:
    scores["kmer"] = scores["kmer"].fillna("NA")

# get alphabet name
if config["alphabet"] in skm.alphabet.ALPHABET_ORDER.keys():
    alphabet_name = skm.alphabet.ALPHABET_ORDER[config["alphabet"]].capitalize()
else:
    alphabet_name = str(config["alphabet"]).capitalize()

# generate [0, 1] labels for binary family assignment
binary_labels = [True if value == family else False for value in data[label]]
le = LabelEncoder()
le.fit(binary_labels)

# set and format input and label arrays; initialize model objs
# need to make training or test set?
X_all = data[f"{family}_score"].values.reshape(-1, 1)
y_all = le.transform(binary_labels).ravel()

# set random seed if specified
rng = np.random.default_rng()
random_state = (
    config["model"]["random_state"]
    if str(config["model"]["random_state"]) != "None"
    else rng.integers(low=0, high=32767)  # max for int16
)

# set and format input and label arrays; initialize model objs
cols = [label, "alphabet_name", "k", "scoring"]
results = {col: [] for col in cols + ["score", "cv_split"]}
X, y = {i: {} for i in range(cv)}, {i: {} for i in range(cv)}
for n in range(cv):

    # remove score cols that were generated from full dataset
    unscored_cols = [col for col in list(data.columns) if "_score" not in col]

    # filter data by training data per CV split
    i_train = data[data[f"train_cv-{n + 1:02d}"]].index
    i_test = data[~data[f"train_cv-{n + 1:02d}"]].index
    df_train = data.iloc[i_train][unscored_cols].reset_index(drop=True)
    df_test = data.iloc[i_test][unscored_cols].reset_index(drop=True)
    df_train_labels = [True if value == family else False for value in df_train[label]]
    df_test_labels = [True if value == family else False for value in df_test[label]]

    # score kmers separately per split
    scorer = skm.score.KmerScorer()
    scorer.fit(
        list(kmer.kmer_set.kmers),
        df_train,
        family,
        label_col=label,
        vec_col="sequence_vector",
        **config["score"]["scaler_kwargs"],
    )

    # append scored sequences to dataframe
    df_train = df_train.merge(
        pd.DataFrame(scorer.scores["sample"]), left_index=True, right_index=True
    )
    if df_train.empty:
        raise ValueError("Blank df")
    df_test = df_test.merge(
        pd.DataFrame(
            scorer.predict(
                skm.utils.to_feature_matrix(df_test["sequence_vector"]),
                list(kmer.kmer_set.kmers),
            )
        ),
        left_index=True,
        right_index=True,
    ).rename(columns={0: f"{family}_score"})

    # save score loadings
    scores = (
        pd.DataFrame(scorer.probabilities, index=scorer.kmers.basis)
        .reset_index()
        .rename(columns={"index": "kmer"})
    )

    # save X,y array data for plot
    X[n]["train"] = df_train[f"{family}_score"].values.reshape(-1, 1)
    y[n]["train"] = le.transform(df_train_labels).ravel()

    X[n]["test"] = df_test[f"{family}_score"].values.reshape(-1, 1)
    y[n]["test"] = le.transform(df_test_labels).ravel()

# ROC-AUC figure
clf = skm.model.SnekmerModel(
    model="logistic",
    params={
        "random_state": random_state,
        "solver": "liblinear",
        "class_weight": "balanced",
    },
)
# clf = LogisticRegression(
#     random_state=random_state, solver="liblinear", class_weight="balanced"
# )
fig, ax, auc_rocs = skm.plot.cv_roc_curve(
    clf, X, y, title=f"{family} ROC Curve ({alphabet_name}, k={config['k']})"
)

# collate ROC-AUC results
results[label] += [family] * cv
results["alphabet_name"] += [alphabet_name.lower()] * cv
results["k"] += [config["k"]] * cv
results["scoring"] += ["auc_roc"] * cv
results["score"] += auc_rocs
results["cv_split"] += [i + 1 for i in range(cv)]

# save ROC-AUC figure
if not exists(snakemake.output.figs):
    makedirs(snakemake.output.figs)
fig.savefig(
    join(
        snakemake.output.figs,
        (f"{family}_roc-auc-curve_{alphabet_name.lower()}" f"_k-{config['k']:02d}.png"),
    ),
    dpi=fig.dpi,
)
fig.clf()
plt.close("all")

# PR-AUC figure
fig, ax, pr_aucs = skm.plot.cv_pr_curve(
    clf, X, y, title=f"{family} PR Curve ({alphabet_name}, k={config['k']})"
)

# collate PR-AUC results
results[label] += [family] * cv
results["alphabet_name"] += [alphabet_name.lower()] * cv
results["k"] += [config["k"]] * cv
results["scoring"] += ["pr_auc"] * cv
results["score"] += pr_aucs
results["cv_split"] += [i + 1 for i in range(cv)]

# save PR-AUC figure
fig.savefig(
    join(
        snakemake.output.figs,
        (f"{family}_aupr-curve_{alphabet_name.lower()}" f"_k-{config['k']:02d}.png"),
    ),
    dpi=fig.dpi,
)
fig.clf()
plt.close("all")

# save model
clf.fit(X_all, y_all)
with open(snakemake.output.model, "wb") as save_model:
    pickle.dump(clf, save_model)

# save full results
pd.DataFrame(results).to_csv(snakemake.output.results, index=False)
